""" nmigen operator functions / utils

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


    eq:
    --

    a strategically very important function that is identical in function
    to nmigen's Signal.eq function, except it may take objects, or a list
    of objects, or a tuple of objects, and where objects may also be
    Records.
"""

from nmigen import Signal, Cat, Value
from nmigen.hdl.ast import ArrayProxy
from nmigen.hdl.rec import Record, Layout

from abc import ABCMeta, abstractmethod
from collections.abc import Sequence, Iterable
import inspect


class Visitor2:
    """ a helper class for iterating twin-argument compound data structures.

        Record is a special (unusual, recursive) case, where the input may be
        specified as a dictionary (which may contain further dictionaries,
        recursively), where the field names of the dictionary must match
        the Record's field spec.  Alternatively, an object with the same
        member names as the Record may be assigned: it does not have to
        *be* a Record.

        ArrayProxy is also special-cased, it's a bit messy: whilst ArrayProxy
        has an eq function, the object being assigned to it (e.g. a python
        object) might not.  despite the *input* having an eq function,
        that doesn't help us, because it's the *ArrayProxy* that's being
        assigned to.  so.... we cheat.  use the ports() function of the
        python object, enumerate them, find out the list of Signals that way,
        and assign them.
    """
    def iterator2(self, o, i):
        if isinstance(o, dict):
            yield from self.dict_iter2(o, i)

        if not isinstance(o, Sequence):
            o, i = [o], [i]
        for (ao, ai) in zip(o, i):
            #print ("visit", fn, ao, ai)
            if isinstance(ao, Record):
                yield from self.record_iter2(ao, ai)
            elif isinstance(ao, ArrayProxy) and not isinstance(ai, Value):
                yield from self.arrayproxy_iter2(ao, ai)
            else:
                yield (ao, ai)

    def dict_iter2(self, o, i):
        for (k, v) in o.items():
            print ("d-iter", v, i[k])
            yield (v, i[k])
        return res

    def _not_quite_working_with_all_unit_tests_record_iter2(self, ao, ai):
        print ("record_iter2", ao, ai, type(ao), type(ai))
        if isinstance(ai, Value):
            if isinstance(ao, Sequence):
                ao, ai = [ao], [ai]
            for o, i in zip(ao, ai):
                yield (o, i)
            return
        for idx, (field_name, field_shape, _) in enumerate(ao.layout):
            if isinstance(field_shape, Layout):
                val = ai.fields
            else:
                val = ai
            if hasattr(val, field_name): # check for attribute
                val = getattr(val, field_name)
            else:
                val = val[field_name] # dictionary-style specification
            yield from self.iterator2(ao.fields[field_name], val)

    def record_iter2(self, ao, ai):
        for idx, (field_name, field_shape, _) in enumerate(ao.layout):
            if isinstance(field_shape, Layout):
                val = ai.fields
            else:
                val = ai
            if hasattr(val, field_name): # check for attribute
                val = getattr(val, field_name)
            else:
                val = val[field_name] # dictionary-style specification
            yield from self.iterator2(ao.fields[field_name], val)

    def arrayproxy_iter2(self, ao, ai):
        #print ("arrayproxy_iter2", ai.ports(), ai, ao)
        for p in ai.ports():
            #print ("arrayproxy - p", p, p.name, ao)
            op = getattr(ao, p.name)
            yield from self.iterator2(op, p)


class Visitor:
    """ a helper class for iterating single-argument compound data structures.
        similar to Visitor2.
    """
    def iterate(self, i):
        """ iterate a compound structure recursively using yield
        """
        if not isinstance(i, Sequence):
            i = [i]
        for ai in i:
            #print ("iterate", ai)
            if isinstance(ai, Record):
                #print ("record", list(ai.layout))
                yield from self.record_iter(ai)
            elif isinstance(ai, ArrayProxy) and not isinstance(ai, Value):
                yield from self.array_iter(ai)
            else:
                yield ai

    def record_iter(self, ai):
        for idx, (field_name, field_shape, _) in enumerate(ai.layout):
            if isinstance(field_shape, Layout):
                val = ai.fields
            else:
                val = ai
            if hasattr(val, field_name): # check for attribute
                val = getattr(val, field_name)
            else:
                val = val[field_name] # dictionary-style specification
            #print ("recidx", idx, field_name, field_shape, val)
            yield from self.iterate(val)

    def array_iter(self, ai):
        for p in ai.ports():
            yield from self.iterate(p)


def eq(o, i):
    """ makes signals equal: a helper routine which identifies if it is being
        passed a list (or tuple) of objects, or signals, or Records, and calls
        the objects' eq function.
    """
    res = []
    for (ao, ai) in Visitor2().iterator2(o, i):
        rres = ao.eq(ai)
        if not isinstance(rres, Sequence):
            rres = [rres]
        res += rres
    return res


def shape(i):
    #print ("shape", i)
    r = 0
    for part in list(i):
        #print ("shape?", part)
        s, _ = part.shape()
        r += s
    return r, False


def cat(i):
    """ flattens a compound structure recursively using Cat
    """
    from nmigen._utils import flatten
    #res = list(flatten(i)) # works (as of nmigen commit f22106e5) HOWEVER...
    res = list(Visitor().iterate(i)) # needed because input may be a sequence
    return Cat(*res)





"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


"""
from nmigen.compat.sim import run_simulation
from nmigen.cli import verilog, rtlil
from nmigen import Record, Signal, Module, Const, Elaboratable, Mux

""" jk latch

module jk(q,q1,j,k,c);
output q,q1;
input j,k,c;
reg q,q1;
initial begin q=1'b0; q1=1'b1; end
always @ (posedge c)
  begin
    case({j,k})
         {1'b0,1'b0}:begin q=q; q1=q1; end
         {1'b0,1'b1}: begin q=1'b0; q1=1'b1; end
         {1'b1,1'b0}:begin q=1'b1; q1=1'b0; end
         {1'b1,1'b1}: begin q=~q; q1=~q1; end
    endcase
   end
endmodule
"""


def latchregister(m, incoming, outgoing, settrue, name=None):
    """latchregister

    based on a conditon, "settrue", incoming data will be "latched"
    into a register and passed out on "outgoing".

    * if "settrue" is ASSERTED, outgoing is COMBINATORIALLY equal to incoming
    * on the same cycle that settrue is DEASSERTED, outgoing REMAINS
      equal (indefinitely) to the incoming value
    """
    # make reg same as input. reset OK.
    if isinstance(incoming, Record):
        reg = Record.like(incoming, name=name)
    else:
        reg = Signal.like(incoming, name=name)
    m.d.comb += outgoing.eq(Mux(settrue, incoming, reg))
    with m.If(settrue): # pass in some kind of expression/condition here
        m.d.sync += reg.eq(incoming)      # latch input into register


def mkname(prefix, suffix):
    if suffix is None:
        return prefix
    return "%s_%s" % (prefix, suffix)


class SRLatch(Elaboratable):
    def __init__(self, sync=True, llen=1, name=None):
        self.sync = sync
        self.llen = llen
        s_n, r_n = mkname("s", name), mkname("r", name)
        q_n, qn_n = mkname("q", name), mkname("qn", name)
        qlq_n = mkname("qlq", name)
        self.s = Signal(llen, name=s_n, reset=0)
        self.r = Signal(llen, name=r_n, reset=(1<<llen)-1) # defaults to off
        self.q = Signal(llen, name=q_n, reset_less=True)
        self.qn = Signal(llen, name=qn_n, reset_less=True)
        self.qlq = Signal(llen, name=qlq_n, reset_less=True)

    def elaborate(self, platform):
        m = Module()
        q_int = Signal(self.llen)

        m.d.sync += q_int.eq((q_int & ~self.r) | self.s)
        if self.sync:
            m.d.comb += self.q.eq(q_int)
        else:
            m.d.comb += self.q.eq((q_int & ~self.r) | self.s)
        m.d.comb += self.qn.eq(~self.q)
        m.d.comb += self.qlq.eq(self.q | q_int) # useful output

        return m

    def ports(self):
        return self.s, self.r, self.q, self.qn


def sr_sim(dut):
    yield dut.s.eq(0)
    yield dut.r.eq(0)
    yield
    yield
    yield
    yield dut.s.eq(1)
    yield
    yield
    yield
    yield dut.s.eq(0)
    yield
    yield
    yield
    yield dut.r.eq(1)
    yield
    yield
    yield
    yield dut.r.eq(0)
    yield
    yield
    yield

def test_sr():
    dut = SRLatch(llen=4)
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_srlatch.il", "w") as f:
        f.write(vl)

    run_simulation(dut, sr_sim(dut), vcd_name='test_srlatch.vcd')

    dut = SRLatch(sync=False, llen=4)
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_srlatch_async.il", "w") as f:
        f.write(vl)

    run_simulation(dut, sr_sim(dut), vcd_name='test_srlatch_async.vcd')

if __name__ == '__main__':
    test_sr()



"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+

"""

# need to ripple the starting LSB of each partition up through the
# rest of the partition.  a Mux on the partition gate therefore selects
# either the current "thing" being propagated, or, if the gate is set open,
# will select the current bit from the input.
#
# this is actually a useful function, it's one of "set before first" or
# "set after first" from vector predication processing.

from nmigen import Signal, Module, Elaboratable, Mux, Cat
from nmigen.cli import main


class RippleLSB(Elaboratable):
    """RippleLSB

    based on a partition mask, the LSB is "rippled" (duplicated)
    up to the beginning of the next partition.
    """
    def __init__(self, width):
        self.width = width
        self.results_in = Signal(width, reset_less=True)
        self.gates = Signal(width-1, reset_less=True)
        self.output = Signal(width, reset_less=True)

    def elaborate(self, platform):
        m = Module()
        comb = m.d.comb
        width = self.width

        current_result = self.results_in[0]
        comb += self.output[0].eq(current_result)

        for i in range(width-1):
            cur = Mux(self.gates[i], self.results_in[i+1], self.output[i])
            comb += self.output[i+1].eq(cur)

        return m


class MoveMSBDown(Elaboratable):
    """MoveMSBDown

    based on a partition mask, moves the MSB down to the LSB position.
    only the MSB is relevant, other bits are ignored.  works by first
    rippling the MSB across the entire partition (TODO: split that out
    into its own useful module), then ANDs the (new) LSB with the
    partition mask to isolate it.
    """
    def __init__(self, width):
        self.width = width
        self.results_in = Signal(width, reset_less=True)
        self.gates = Signal(width-1, reset_less=True)
        self.output = Signal(width, reset_less=True)

    def elaborate(self, platform):
        m = Module()
        comb = m.d.comb
        width = self.width
        intermed = Signal(width, reset_less=True)

        # first propagate MSB down until the nearest partition gate
        comb += intermed[-1].eq(self.results_in[-1]) # start at MSB
        for i in range(width-2, -1, -1):
            cur = Mux(self.gates[i], self.results_in[i], intermed[i+1])
            comb += intermed[i].eq(cur)

        # now only select those bits where the mask starts
        out = [intermed[0]] # LSB of first part always set
        for i in range(width-1): # length of partition gates
            out.append(self.gates[i] & intermed[i+1])
        comb += self.output.eq(Cat(*out))

        return m


if __name__ == "__main__":
    # python3 ieee754/part_cmp/ripple.py generate -t il > ripple.il
    # then check with yosys "read_ilang ripple.il; show top"
    alu = MoveMSBDown(width=4)
    main(alu, ports=[alu.results_in, alu.gates, alu.output])




# based on ariane plru, from tlb.sv

from nmigen import Signal, Module, Cat, Const, Repl
from nmigen.hdl.ir import Elaboratable
from nmigen.cli import rtlil
from nmigen.utils import log2_int


class PLRU(Elaboratable):
    """ PLRU - Pseudo Least Recently Used Replacement

        PLRU-tree indexing:
        lvl0        0
                   / \
                  /   \
        lvl1     1     2
                / \   / \
        lvl2   3   4 5   6
              / \ /\/\  /\
             ... ... ... ...
    """

    def __init__(self, BITS):
        self.BITS = BITS
        self.acc_en = Signal(BITS)
        self.acc_i = Signal()
        self.lru_o = Signal(BITS)

    def elaborate(self, platform=None):
        m = Module()

        # Tree (bit per entry)
        TLBSZ = 2*(self.BITS-1)
        plru_tree = Signal(TLBSZ)

        # Just predefine which nodes will be set/cleared
        # E.g. for a TLB with 8 entries, the for-loop is semantically
        # equivalent to the following pseudo-code:
        # unique case (1'b1)
        # acc_en[7]: plru_tree[0, 2, 6] = {1, 1, 1};
        # acc_en[6]: plru_tree[0, 2, 6] = {1, 1, 0};
        # acc_en[5]: plru_tree[0, 2, 5] = {1, 0, 1};
        # acc_en[4]: plru_tree[0, 2, 5] = {1, 0, 0};
        # acc_en[3]: plru_tree[0, 1, 4] = {0, 1, 1};
        # acc_en[2]: plru_tree[0, 1, 4] = {0, 1, 0};
        # acc_en[1]: plru_tree[0, 1, 3] = {0, 0, 1};
        # acc_en[0]: plru_tree[0, 1, 3] = {0, 0, 0};
        # default: begin /* No hit */ end
        # endcase

        LOG_TLB = log2_int(self.BITS, False)
        hit = Signal(self.BITS, reset_less=True)
        m.d.comb += hit.eq(Repl(self.acc_i, self.BITS) & self.acc_en)

        for i in range(self.BITS):
            # we got a hit so update the pointer as it was least recently used
            with m.If(hit[i]):
                # Set the nodes to the values we would expect
                for lvl in range(LOG_TLB):
                    idx_base = (1 << lvl)-1
                    # lvl0 <=> MSB, lvl1 <=> MSB-1, ...
                    shift = LOG_TLB - lvl
                    new_idx = Const(~((i >> (shift-1)) & 1), 1)
                    plru_idx = idx_base + (i >> shift)
                    #print("plru", i, lvl, hex(idx_base),
                    #      plru_idx, shift, new_idx)
                    m.d.sync += plru_tree[plru_idx].eq(new_idx)

        # Decode tree to write enable signals
        # Next for-loop basically creates the following logic for e.g.
        # an 8 entry TLB (note: pseudo-code obviously):
        # replace_en[7] = &plru_tree[ 6, 2, 0]; #plru_tree[0,2,6]=={1,1,1}
        # replace_en[6] = &plru_tree[~6, 2, 0]; #plru_tree[0,2,6]=={1,1,0}
        # replace_en[5] = &plru_tree[ 5,~2, 0]; #plru_tree[0,2,5]=={1,0,1}
        # replace_en[4] = &plru_tree[~5,~2, 0]; #plru_tree[0,2,5]=={1,0,0}
        # replace_en[3] = &plru_tree[ 4, 1,~0]; #plru_tree[0,1,4]=={0,1,1}
        # replace_en[2] = &plru_tree[~4, 1,~0]; #plru_tree[0,1,4]=={0,1,0}
        # replace_en[1] = &plru_tree[ 3,~1,~0]; #plru_tree[0,1,3]=={0,0,1}
        # replace_en[0] = &plru_tree[~3,~1,~0]; #plru_tree[0,1,3]=={0,0,0}
        # For each entry traverse the tree. If every tree-node matches
        # the corresponding bit of the entry's index, this is
        # the next entry to replace.
        replace = []
        for i in range(self.BITS):
            en = []
            for lvl in range(LOG_TLB):
                idx_base = (1 << lvl)-1
                # lvl0 <=> MSB, lvl1 <=> MSB-1, ...
                shift = LOG_TLB - lvl
                new_idx = (i >> (shift-1)) & 1
                plru_idx = idx_base + (i >> shift)
                plru = Signal(reset_less=True,
                              name="plru-%d-%d-%d-%d" % \
                                    (i, lvl, plru_idx, new_idx))
                m.d.comb += plru.eq(plru_tree[plru_idx])
                if new_idx:
                    en.append(~plru)  # yes inverted (using bool() below)
                else:
                    en.append(plru)  # yes inverted (using bool() below)
            #print("plru", i, en)
            # boolean logic manipulation:
            # plru0 & plru1 & plru2 == ~(~plru0 | ~plru1 | ~plru2)
            replace.append(~Cat(*en).bool())
        m.d.comb += self.lru_o.eq(Cat(*replace))

        return m

    def ports(self):
        return [self.acc_en, self.lru_o, self.acc_i]


if __name__ == '__main__':
    dut = PLRU(8)
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_plru.il", "w") as f:
        f.write(vl)




from nmigen import Module, Signal, Elaboratable, Cat, Repl
import math
""" This module is much more efficient than PriorityEncoder
    although it is functionally identical.
    see https://bugs.libre-soc.org/show_bug.cgi?id=326

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+

"""

class CLZ(Elaboratable):
    def __init__(self, width):
        self.width = width
        self.sig_in = Signal(width, reset_less=True)
        out_width = math.ceil(math.log2(width+1))
        self.lz = Signal(out_width)

    def generate_pairs(self, m):
        comb = m.d.comb
        pairs = []
        for i in range(0, self.width, 2):
            if i+1 >= self.width:
                pair = Signal(1, name="cnt_1_%d" % (i/2))
                comb += pair.eq(~self.sig_in[i])
                pairs.append((pair, 1))
            else:
                pair = Signal(2, name="pair%d" % i)
                comb += pair.eq(self.sig_in[i:i+2])

                pair_cnt = Signal(2, name="cnt_1_%d" % (i/2))
                with m.Switch(pair):
                    with m.Case(0):
                        comb += pair_cnt.eq(2)
                    with m.Case(1):
                        comb += pair_cnt.eq(1)
                    with m.Default():
                        comb += pair_cnt.eq(0)
                pairs.append((pair_cnt, 2))  # append pair, max_value
        return pairs

    def combine_pairs(self, m, iteration, pairs):
        comb = m.d.comb
        length = len(pairs)
        ret = []
        for i in range(0, length, 2):
            if i+1 >= length:
                right, mv = pairs[i]
                width = right.width
                new_pair = Signal(width, name="cnt_%d_%d" % (iteration, i))
                comb += new_pair.eq(Cat(right, 0))
                ret.append((new_pair, mv))
            else:
                left, lv = pairs[i+1]
                right, rv = pairs[i]
                width = right.width + 1
                new_pair = Signal(width, name="cnt_%d_%d" % (iteration, i))
                if rv == lv:
                    with m.If(left[-1] == 1):
                        with m.If(right[-1] == 1):
                            comb += new_pair.eq(Cat(Repl(0, width-1), 1))
                        with m.Else():
                            comb += new_pair.eq(Cat(right[0:-1], 0b01))
                    with m.Else():
                        comb += new_pair.eq(Cat(left, 0))
                else:
                    with m.If(left == lv):
                        comb += new_pair.eq(right + left)
                    with m.Else():
                        comb += new_pair.eq(left)

                ret.append((new_pair, lv+rv))
        return ret

    def elaborate(self, platform):
        m = Module()
        comb = m.d.comb

        pairs = self.generate_pairs(m)
        i = 2
        while len(pairs) > 1:
            pairs = self.combine_pairs(m, i, pairs)
            i += 1

        comb += self.lz.eq(pairs[0][0])

        return m




"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


"""

""" Example 5: Making use of PyRTL and Introspection. """

from collections.abc import Sequence

from nmigen import Signal
from nmigen.hdl.rec import Record
from nmigen import tracer
from nmigen.compat.fhdl.bitcontainer import value_bits_sign
from contextlib import contextmanager

from nmutil.nmoperator import eq
from nmutil.singlepipe import StageCls, ControlBase, BufferedHandshake
from nmutil.singlepipe import UnbufferedPipeline


# The following example shows how pyrtl can be used to make some interesting
# hardware structures using python introspection.  In particular, this example
# makes a N-stage pipeline structure.  Any specific pipeline is then a derived
# class of SimplePipeline where methods with names starting with "stage" are
# stages, and new members with names not starting with "_" are to be registered
# for the next stage.

def like(value, rname, pipe, pipemode=False):
    if isinstance(value, ObjectProxy):
        return ObjectProxy.like(pipe, value, pipemode=pipemode,
                                name=rname, reset_less=True)
    else:
        return Signal(value_bits_sign(value), name=rname,
                             reset_less=True)
        return Signal.like(value, name=rname, reset_less=True)

def get_assigns(_assigns):
    assigns = []
    for e in _assigns:
        if isinstance(e, ObjectProxy):
            assigns += get_assigns(e._assigns)
        else:
            assigns.append(e)
    return assigns


def get_eqs(_eqs):
    eqs = []
    for e in _eqs:
        if isinstance(e, ObjectProxy):
            eqs += get_eqs(e._eqs)
        else:
            eqs.append(e)
    return eqs


class ObjectProxy:
    def __init__(self, m, name=None, pipemode=False, syncmode=True):
        self._m = m
        if name is None:
            name = tracer.get_var_name(default=None)
        self.name = name
        self._pipemode = pipemode
        self._syncmode = syncmode
        self._eqs = {}
        self._assigns = []
        self._preg_map = {}

    @classmethod
    def like(cls, m, value, pipemode=False, name=None, src_loc_at=0, **kwargs):
        name = name or tracer.get_var_name(depth=2 + src_loc_at,
                                            default="$like")

        src_loc_at_1 = 1 + src_loc_at
        r = ObjectProxy(m, value.name, pipemode)
        #for a, aname in value._preg_map.items():
        #    r._preg_map[aname] = like(a, aname, m, pipemode)
        for a in value.ports():
            aname = a.name
            r._preg_map[aname] = like(a, aname, m, pipemode)
        return r

    def __repr__(self):
        subobjs = []
        for a in self.ports():
            aname = a.name
            ai = self._preg_map[aname]
            subobjs.append(repr(ai))
        return "<OP %s>" % subobjs

    def get_specs(self, liked=False):
        res = []
        for k, v in self._preg_map.items():
            #v = like(v, k, stage._m)
            res.append(v)
            if isinstance(v, ObjectProxy):
                res += v.get_specs()
        return res

    def eq(self, i):
        print ("ObjectProxy eq", self, i)
        res = []
        for a in self.ports():
            aname = a.name
            ai = i._preg_map[aname]
            res.append(a.eq(ai))
        return res

    def ports(self):
        res = []
        for aname, a in self._preg_map.items():
            if isinstance(a, Signal) or isinstance(a, ObjectProxy) or \
               isinstance(a, Record):
                res.append(a)
        #print ("ObjectPorts", res)
        return res

    def __getattr__(self, name):
        try:
            v = self._preg_map[name]
            return v
            #return like(v, name, self._m)
        except KeyError:
            raise AttributeError(
                'error, no pipeline register "%s" defined for OP %s'
                % (name, self.name))

    def __setattr__(self, name, value):
        if name.startswith('_') or name in ['name', 'ports', 'eq', 'like']:
            # do not do anything tricky with variables starting with '_'
            object.__setattr__(self, name, value)
            return
        #rname = "%s_%s" % (self.name, name)
        rname = name
        new_pipereg = like(value, rname, self._m, self._pipemode)
        self._preg_map[name] = new_pipereg
        #object.__setattr__(self, name, new_pipereg)
        if self._pipemode:
            #print ("OP pipemode", self._syncmode, new_pipereg, value)
            assign = eq(new_pipereg, value)
            if self._syncmode:
                self._m.d.sync += assign
            else:
                self._m.d.comb += assign
        elif self._m:
            #print ("OP !pipemode assign", new_pipereg, value, type(value))
            self._m.d.comb += eq(new_pipereg, value)
        else:
            #print ("OP !pipemode !m", new_pipereg, value, type(value))
            self._assigns += eq(new_pipereg, value)
            if isinstance(value, ObjectProxy):
                #print ("OP, defer assigns:", value._assigns)
                self._assigns += value._assigns
                self._eqs.append(value._eqs)


class PipelineStage:
    """ Pipeline builder stage with auto generation of pipeline registers.
    """

    def __init__(self, name, m, prev=None, pipemode=False, ispec=None):
        self._m = m
        self._stagename = name
        self._preg_map = {'__nextstage__': {}}
        self._prev_stage = prev
        self._ispec = ispec
        if ispec:
            self._preg_map[self._stagename] = ispec
        if prev:
            print ("prev", prev._stagename, prev._preg_map)
            #if prev._stagename in prev._preg_map:
            #    m = prev._preg_map[prev._stagename]
            #    self._preg_map[prev._stagename] = m
            if '__nextstage__' in prev._preg_map:
                m = prev._preg_map['__nextstage__']
                m = likedict(m)
                self._preg_map[self._stagename] = m
                #for k, v in m.items():
                    #m[k] = like(v, k, self._m)
                print ("make current", self._stagename, m)
        self._pipemode = pipemode
        self._eqs = {}
        self._assigns = []

    def __getattribute__(self, name):
        if name.startswith('_'):
            return object.__getattribute__(self, name)
        #if name in self._preg_map['__nextstage__']:
        #    return self._preg_map['__nextstage__'][name]
        try:
            print ("getattr", name, object.__getattribute__(self, '_preg_map'))
            v = self._preg_map[self._stagename][name]
            return v
            #return like(v, name, self._m)
        except KeyError:
            raise AttributeError(
                'error, no pipeline register "%s" defined for stage %s'
                % (name, self._stagename))

    def __setattr__(self, name, value):
        if name.startswith('_'):
            # do not do anything tricky with variables starting with '_'
            object.__setattr__(self, name, value)
            return
        pipereg_id = self._stagename
        rname = 'pipereg_' + pipereg_id + '_' + name
        new_pipereg = like(value, rname, self._m, self._pipemode)
        next_stage = '__nextstage__'
        if next_stage not in self._preg_map:
            self._preg_map[next_stage] = {}
        self._preg_map[next_stage][name] = new_pipereg
        print ("setattr", name, value, self._preg_map)
        if self._pipemode:
            self._eqs[name] = new_pipereg
            assign = eq(new_pipereg, value)
            print ("pipemode: append", new_pipereg, value, assign)
            if isinstance(value, ObjectProxy):
                print ("OP, assigns:", value._assigns)
                self._assigns += value._assigns
                self._eqs[name]._eqs = value._eqs
            #self._m.d.comb += assign
            self._assigns += assign
        elif self._m:
            print ("!pipemode: assign", new_pipereg, value)
            assign = eq(new_pipereg, value)
            self._m.d.sync += assign
        else:
            print ("!pipemode !m: defer assign", new_pipereg, value)
            assign = eq(new_pipereg, value)
            self._eqs[name] = new_pipereg
            self._assigns += assign
            if isinstance(value, ObjectProxy):
                print ("OP, defer assigns:", value._assigns)
                self._assigns += value._assigns
                self._eqs[name]._eqs = value._eqs

def likelist(specs):
    res = []
    for v in specs:
        res.append(like(v, v.name, None, pipemode=True))
    return res

def likedict(specs):
    if not isinstance(specs, dict):
        return like(specs, specs.name, None, pipemode=True)
    res = {}
    for k, v in specs.items():
        res[k] = likedict(v)
    return res


class AutoStage(StageCls):
    def __init__(self, inspecs, outspecs, eqs, assigns):
        self.inspecs, self.outspecs = inspecs, outspecs
        self.eqs, self.assigns = eqs, assigns
        #self.o = self.ospec()
    def ispec(self): return likedict(self.inspecs)
    def ospec(self): return likedict(self.outspecs)

    def process(self, i):
        print ("stage process", i)
        return self.eqs

    def setup(self, m, i):
        print ("stage setup i", i, m)
        print ("stage setup inspecs", self.inspecs)
        print ("stage setup outspecs", self.outspecs)
        print ("stage setup eqs", self.eqs)
        #self.o = self.ospec()
        m.d.comb += eq(self.inspecs, i)
        #m.d.comb += eq(self.outspecs, self.eqs)
        #m.d.comb += eq(self.o, i)


class AutoPipe(UnbufferedPipeline):
    def __init__(self, stage, assigns):
        UnbufferedPipeline.__init__(self, stage)
        self.assigns = assigns

    def elaborate(self, platform):
        m = UnbufferedPipeline.elaborate(self, platform)
        m.d.comb += self.assigns
        print ("assigns", self.assigns, m)
        return m


class PipeManager:
    def __init__(self, m, pipemode=False, pipetype=None):
        self.m = m
        self.pipemode = pipemode
        self.pipetype = pipetype

    @contextmanager
    def Stage(self, name, prev=None, ispec=None):
        if ispec:
            ispec = likedict(ispec)
        print ("start stage", name, ispec)
        stage = PipelineStage(name, None, prev, self.pipemode, ispec=ispec)
        try:
            yield stage, self.m #stage._m
        finally:
            pass
        if self.pipemode:
            if stage._ispec:
                print ("use ispec", stage._ispec)
                inspecs = stage._ispec
            else:
                inspecs = self.get_specs(stage, name)
                #inspecs = likedict(inspecs)
            outspecs = self.get_specs(stage, '__nextstage__', liked=True)
            print ("stage inspecs", name, inspecs)
            print ("stage outspecs", name, outspecs)
            eqs = stage._eqs # get_eqs(stage._eqs)
            assigns = get_assigns(stage._assigns)
            print ("stage eqs", name, eqs)
            print ("stage assigns", name, assigns)
            s = AutoStage(inspecs, outspecs, eqs, assigns)
            self.stages.append(s)
        print ("end stage", name, self.pipemode, "\n")

    def get_specs(self, stage, name, liked=False):
        return stage._preg_map[name]
        if name in stage._preg_map:
            res = []
            for k, v in stage._preg_map[name].items():
                #v = like(v, k, stage._m)
                res.append(v)
                #if isinstance(v, ObjectProxy):
                #    res += v.get_specs()
            return res
        return {}

    def __enter__(self):
        self.stages = []
        return self

    def __exit__(self, *args):
        print ("exit stage", args)
        pipes = []
        cb = ControlBase()
        for s in self.stages:
            print ("stage specs", s, s.inspecs, s.outspecs)
            if self.pipetype == 'buffered':
                p = BufferedHandshake(s)
            else:
                p = AutoPipe(s, s.assigns)
            pipes.append(p)
            self.m.submodules += p

        self.m.d.comb += cb.connect(pipes)


class SimplePipeline:
    """ Pipeline builder with auto generation of pipeline registers.
    """

    def __init__(self, m):
        self._m = m
        self._pipeline_register_map = {}
        self._current_stage_num = 0

    def _setup(self):
        stage_list = []
        for method in dir(self):
            if method.startswith('stage'):
                stage_list.append(method)
        for stage in sorted(stage_list):
            stage_method = getattr(self, stage)
            stage_method()
            self._current_stage_num += 1

    def __getattr__(self, name):
        try:
            return self._pipeline_register_map[self._current_stage_num][name]
        except KeyError:
            raise AttributeError(
                'error, no pipeline register "%s" defined for stage %d'
                % (name, self._current_stage_num))

    def __setattr__(self, name, value):
        if name.startswith('_'):
            # do not do anything tricky with variables starting with '_'
            object.__setattr__(self, name, value)
            return
        next_stage = self._current_stage_num + 1
        pipereg_id = str(self._current_stage_num) + 'to' + str(next_stage)
        rname = 'pipereg_' + pipereg_id + '_' + name
        #new_pipereg = Signal(value_bits_sign(value), name=rname,
        #                     reset_less=True)
        if isinstance(value, ObjectProxy):
            new_pipereg = ObjectProxy.like(self._m, value,
                                           name=rname, reset_less = True)
        else:
            new_pipereg = Signal.like(value, name=rname, reset_less = True)
        if next_stage not in self._pipeline_register_map:
            self._pipeline_register_map[next_stage] = {}
        self._pipeline_register_map[next_stage][name] = new_pipereg
        self._m.d.sync += eq(new_pipereg, value)




from nmigen import Module, Signal, Elaboratable, Mux, Const
from nmigen.asserts import Assert, AnyConst, Assume
from nmutil.formaltest import FHDLTestCase
from nmigen.cli import rtlil

from nmutil.clz import CLZ
import unittest


# This defines a module to drive the device under test and assert
# properties about its outputs
class Driver(Elaboratable):
    def __init__(self):
        # inputs and outputs
        pass

    def elaborate(self, platform):
        m = Module()
        comb = m.d.comb
        width = 10

        m.submodules.dut = dut = CLZ(width)
        sig_in = Signal.like(dut.sig_in)
        count = Signal.like(dut.lz)


        m.d.comb += [
            sig_in.eq(AnyConst(width)),
            dut.sig_in.eq(sig_in),
            count.eq(dut.lz)]

        result = Const(width)
        for i in range(width):
            print(result)
            result_next = Signal.like(count, name="count_%d" % i)
            with m.If(sig_in[i] == 1):
                comb += result_next.eq(width-i-1)
            with m.Else():
                comb += result_next.eq(result)
            result = result_next

        result_sig = Signal.like(count)
        comb += result_sig.eq(result)

        comb += Assert(result_sig == count)
        
        # setup the inputs and outputs of the DUT as anyconst

        return m

class CLZTestCase(FHDLTestCase):
    def test_proof(self):
        module = Driver()
        self.assertFormal(module, mode="bmc", depth=4)
    def test_ilang(self):
        dut = Driver()
        vl = rtlil.convert(dut, ports=[])
        with open("clz.il", "w") as f:
            f.write(vl)

if __name__ == '__main__':
    unittest.main()



from nmigen import Module, Signal
from nmigen.back.pysim import Simulator, Delay

from nmutil.clz import CLZ
import unittest
import math
import random


class CLZTestCase(unittest.TestCase):
    def run_test(self, inputs, width=8):

        m = Module()

        m.submodules.dut = dut = CLZ(width)
        sig_in = Signal.like(dut.sig_in)
        count = Signal.like(dut.lz)


        m.d.comb += [
            dut.sig_in.eq(sig_in),
            count.eq(dut.lz)]

        sim = Simulator(m)

        def process():
            for i in inputs:
                yield sig_in.eq(i)
                yield Delay(1e-6)
        sim.add_process(process)
        with sim.write_vcd("clz.vcd", "clz.gtkw", traces=[
                sig_in, count]):
            sim.run()

    def test_selected(self):
        inputs = [0, 15, 10, 127]
        self.run_test(iter(inputs), width=8)

    def test_non_power_2(self):
        inputs = [0, 128, 512]
        self.run_test(iter(inputs), width=11)


if __name__ == "__main__":
    unittest.main()



class StepLimiter:
    def __init__(self, limit=100000):
        self.limit = limit
        self.step_count = 0
        assert self.step_count < self.limit, "step count limit reached"

    def step(self):
        self.step_count += 1
        assert self.step_count < self.limit, "step count limit reached"

    def __iter__(self):
        return self

    def __next__(self):
        self.step()
        return None



""" Pipeline API.  For multi-input and multi-output variants, see multipipe.

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


    Associated development bugs:
    * http://bugs.libre-riscv.org/show_bug.cgi?id=148
    * http://bugs.libre-riscv.org/show_bug.cgi?id=64
    * http://bugs.libre-riscv.org/show_bug.cgi?id=57

    Important: see Stage API (stageapi.py) and IO Control API
    (iocontrol.py) in combination with below.  This module
    "combines" the Stage API with the IO Control API to create
    the Pipeline API.

    The one critically important key difference between StageAPI and
    PipelineAPI:

        * StageAPI: combinatorial (NO REGISTERS / LATCHES PERMITTED)
        * PipelineAPI: synchronous registers / latches get added here

    RecordBasedStage:
    ----------------

    A convenience class that takes an input shape, output shape, a
    "processing" function and an optional "setup" function.  Honestly
    though, there's not much more effort to just... create a class
    that returns a couple of Records (see ExampleAddRecordStage in
    examples).

    PassThroughStage:
    ----------------

    A convenience class that takes a single function as a parameter,
    that is chain-called to create the exact same input and output spec.
    It has a process() function that simply returns its input.

    Instances of this class are completely redundant if handed to
    StageChain, however when passed to UnbufferedPipeline they
    can be used to introduce a single clock delay.

    ControlBase:
    -----------

    The base class for pipelines.  Contains previous and next ready/valid/data.
    Also has an extremely useful "connect" function that can be used to
    connect a chain of pipelines and present the exact same prev/next
    ready/valid/data API.

    Note: pipelines basically do not become pipelines as such until
    handed to a derivative of ControlBase.  ControlBase itself is *not*
    strictly considered a pipeline class.  Wishbone and AXI4 (master or
    slave) could be derived from ControlBase, for example.
    UnbufferedPipeline:
    ------------------

    A simple stalling clock-synchronised pipeline that has no buffering
    (unlike BufferedHandshake).  Data flows on *every* clock cycle when
    the conditions are right (this is nominally when the input is valid
    and the output is ready).

    A stall anywhere along the line will result in a stall back-propagating
    down the entire chain.  The BufferedHandshake by contrast will buffer
    incoming data, allowing previous stages one clock cycle's grace before
    also having to stall.

    An advantage of the UnbufferedPipeline over the Buffered one is
    that the amount of logic needed (number of gates) is greatly
    reduced (no second set of buffers basically)

    The disadvantage of the UnbufferedPipeline is that the valid/ready
    logic, if chained together, is *combinatorial*, resulting in
    progressively larger gate delay.

    PassThroughHandshake:
    ------------------

    A Control class that introduces a single clock delay, passing its
    data through unaltered.  Unlike RegisterPipeline (which relies
    on UnbufferedPipeline and PassThroughStage) it handles ready/valid
    itself.

    RegisterPipeline:
    ----------------

    A convenience class that, because UnbufferedPipeline introduces a single
    clock delay, when its stage is a PassThroughStage, it results in a Pipeline
    stage that, duh, delays its (unmodified) input by one clock cycle.

    BufferedHandshake:
    ----------------

    nmigen implementation of buffered pipeline stage, based on zipcpu:
    https://zipcpu.com/blog/2017/08/14/strategies-for-pipelining.html

    this module requires quite a bit of thought to understand how it works
    (and why it is needed in the first place).  reading the above is
    *strongly* recommended.

    unlike john dawson's IEEE754 FPU STB/ACK signalling, which requires
    the STB / ACK signals to raise and lower (on separate clocks) before
    data may proceeed (thus only allowing one piece of data to proceed
    on *ALTERNATE* cycles), the signalling here is a true pipeline
    where data will flow on *every* clock when the conditions are right.

    input acceptance conditions are when:
        * incoming previous-stage strobe (p.valid_i) is HIGH
        * outgoing previous-stage ready   (p.ready_o) is LOW

    output transmission conditions are when:
        * outgoing next-stage strobe (n.valid_o) is HIGH
        * outgoing next-stage ready   (n.ready_i) is LOW

    the tricky bit is when the input has valid data and the output is not
    ready to accept it.  if it wasn't for the clock synchronisation, it
    would be possible to tell the input "hey don't send that data, we're
    not ready".  unfortunately, it's not possible to "change the past":
    the previous stage *has no choice* but to pass on its data.

    therefore, the incoming data *must* be accepted - and stored: that
    is the responsibility / contract that this stage *must* accept.
    on the same clock, it's possible to tell the input that it must
    not send any more data.  this is the "stall" condition.

    we now effectively have *two* possible pieces of data to "choose" from:
    the buffered data, and the incoming data.  the decision as to which
    to process and output is based on whether we are in "stall" or not.
    i.e. when the next stage is no longer ready, the output comes from
    the buffer if a stall had previously occurred, otherwise it comes
    direct from processing the input.

    this allows us to respect a synchronous "travelling STB" with what
    dan calls a "buffered handshake".

    it's quite a complex state machine!

    SimpleHandshake
    ---------------

    Synchronised pipeline, Based on:
    https://github.com/ZipCPU/dbgbus/blob/master/hexbus/rtl/hbdeword.v
"""

from nmigen import Signal, Mux, Module, Elaboratable, Const
from nmigen.cli import verilog, rtlil
from nmigen.hdl.rec import Record

from nmutil.queue import Queue
import inspect

from nmutil.iocontrol import (PrevControl, NextControl, Object, RecordObject)
from nmutil.stageapi import (_spec, StageCls, Stage, StageChain, StageHelper)
from nmutil import nmoperator


class RecordBasedStage(Stage):
    """ convenience class which provides a Records-based layout.
        honestly it's a lot easier just to create a direct Records-based
        class (see ExampleAddRecordStage)
    """
    def __init__(self, in_shape, out_shape, processfn, setupfn=None):
        self.in_shape = in_shape
        self.out_shape = out_shape
        self.__process = processfn
        self.__setup = setupfn
    def ispec(self): return Record(self.in_shape)
    def ospec(self): return Record(self.out_shape)
    def process(seif, i): return self.__process(i)
    def setup(seif, m, i): return self.__setup(m, i)


class PassThroughStage(StageCls):
    """ a pass-through stage with its input data spec identical to its output,
        and "passes through" its data from input to output (does nothing).

        use this basically to explicitly make any data spec Stage-compliant.
        (many APIs would potentially use a static "wrap" method in e.g.
         StageCls to achieve a similar effect)
    """
    def __init__(self, iospecfn): self.iospecfn = iospecfn
    def ispec(self): return self.iospecfn()
    def ospec(self): return self.iospecfn()


class ControlBase(StageHelper, Elaboratable):
    """ Common functions for Pipeline API.  Note: a "pipeline stage" only
        exists (conceptually) when a ControlBase derivative is handed
        a Stage (combinatorial block)

        NOTE: ControlBase derives from StageHelper, making it accidentally
        compliant with the Stage API.  Using those functions directly
        *BYPASSES* a ControlBase instance ready/valid signalling, which
        clearly should not be done without a really, really good reason.
    """
    def __init__(self, stage=None, in_multi=None, stage_ctl=False, maskwid=0):
        """ Base class containing ready/valid/data to previous and next stages

            * p: contains ready/valid to the previous stage
            * n: contains ready/valid to the next stage

            Except when calling Controlbase.connect(), user must also:
            * add data_i member to PrevControl (p) and
            * add data_o member to NextControl (n)
            Calling ControlBase._new_data is a good way to do that.
        """
        print ("ControlBase", self, stage, in_multi, stage_ctl)
        StageHelper.__init__(self, stage)

        # set up input and output IO ACK (prev/next ready/valid)
        self.p = PrevControl(in_multi, stage_ctl, maskwid=maskwid)
        self.n = NextControl(stage_ctl, maskwid=maskwid)

        # set up the input and output data
        if stage is not None:
            self._new_data("data")

    def _new_data(self, name):
        """ allocates new data_i and data_o
        """
        self.p.data_i, self.n.data_o = self.new_specs(name)

    @property
    def data_r(self):
        return self.process(self.p.data_i)

    def connect_to_next(self, nxt):
        """ helper function to connect to the next stage data/valid/ready.
        """
        return self.n.connect_to_next(nxt.p)

    def _connect_in(self, prev):
        """ internal helper function to connect stage to an input source.
            do not use to connect stage-to-stage!
        """
        return self.p._connect_in(prev.p)

    def _connect_out(self, nxt):
        """ internal helper function to connect stage to an output source.
            do not use to connect stage-to-stage!
        """
        return self.n._connect_out(nxt.n)

    def connect(self, pipechain):
        """ connects a chain (list) of Pipeline instances together and
            links them to this ControlBase instance:

                      in <----> self <---> out
                       |                   ^
                       v                   |
                    [pipe1, pipe2, pipe3, pipe4]
                       |    ^  |    ^  |     ^
                       v    |  v    |  v     |
                     out---in out--in out---in

            Also takes care of allocating data_i/data_o, by looking up
            the data spec for each end of the pipechain.  i.e It is NOT
            necessary to allocate self.p.data_i or self.n.data_o manually:
            this is handled AUTOMATICALLY, here.

            Basically this function is the direct equivalent of StageChain,
            except that unlike StageChain, the Pipeline logic is followed.

            Just as StageChain presents an object that conforms to the
            Stage API from a list of objects that also conform to the
            Stage API, an object that calls this Pipeline connect function
            has the exact same pipeline API as the list of pipline objects
            it is called with.

            Thus it becomes possible to build up larger chains recursively.
            More complex chains (multi-input, multi-output) will have to be
            done manually.

            Argument:

            * :pipechain: - a sequence of ControlBase-derived classes
                            (must be one or more in length)

            Returns:

            * a list of eq assignments that will need to be added in
              an elaborate() to m.d.comb
        """
        assert len(pipechain) > 0, "pipechain must be non-zero length"
        assert self.stage is None, "do not use connect with a stage"
        eqs = [] # collated list of assignment statements

        # connect inter-chain
        for i in range(len(pipechain)-1):
            pipe1 = pipechain[i]                # earlier
            pipe2 = pipechain[i+1]              # later (by 1)
            eqs += pipe1.connect_to_next(pipe2) # earlier n to later p

        # connect front and back of chain to ourselves
        front = pipechain[0]                # first in chain
        end = pipechain[-1]                 # last in chain
        self.set_specs(front, end) # sets up ispec/ospec functions
        self._new_data("chain") # NOTE: REPLACES existing data
        eqs += front._connect_in(self)      # front p to our p
        eqs += end._connect_out(self)       # end n   to our n

        return eqs

    def set_input(self, i):
        """ helper function to set the input data (used in unit tests)
        """
        return nmoperator.eq(self.p.data_i, i)

    def __iter__(self):
        yield from self.p # yields ready/valid/data (data also gets yielded)
        yield from self.n # ditto

    def ports(self):
        return list(self)

    def elaborate(self, platform):
        """ handles case where stage has dynamic ready/valid functions
        """
        m = Module()
        m.submodules.p = self.p
        m.submodules.n = self.n

        self.setup(m, self.p.data_i)

        if not self.p.stage_ctl:
            return m

        # intercept the previous (outgoing) "ready", combine with stage ready
        m.d.comb += self.p.s_ready_o.eq(self.p._ready_o & self.stage.d_ready)

        # intercept the next (incoming) "ready" and combine it with data valid
        sdv = self.stage.d_valid(self.n.ready_i)
        m.d.comb += self.n.d_valid.eq(self.n.ready_i & sdv)

        return m


class BufferedHandshake(ControlBase):
    """ buffered pipeline stage.  data and strobe signals travel in sync.
        if ever the input is ready and the output is not, processed data
        is shunted in a temporary register.

        Argument: stage.  see Stage API above

        stage-1   p.valid_i >>in   stage   n.valid_o out>>   stage+1
        stage-1   p.ready_o <<out  stage   n.ready_i <<in    stage+1
        stage-1   p.data_i  >>in   stage   n.data_o  out>>   stage+1
                              |             |
                            process --->----^
                              |             |
                              +-- r_data ->-+

        input data p.data_i is read (only), is processed and goes into an
        intermediate result store [process()].  this is updated combinatorially.

        in a non-stall condition, the intermediate result will go into the
        output (update_output).  however if ever there is a stall, it goes
        into r_data instead [update_buffer()].

        when the non-stall condition is released, r_data is the first
        to be transferred to the output [flush_buffer()], and the stall
        condition cleared.

        on the next cycle (as long as stall is not raised again) the
        input may begin to be processed and transferred directly to output.
    """

    def elaborate(self, platform):
        self.m = ControlBase.elaborate(self, platform)

        result = _spec(self.stage.ospec, "r_tmp")
        r_data = _spec(self.stage.ospec, "r_data")

        # establish some combinatorial temporaries
        o_n_validn = Signal(reset_less=True)
        n_ready_i = Signal(reset_less=True, name="n_i_rdy_data")
        nir_por = Signal(reset_less=True)
        nir_por_n = Signal(reset_less=True)
        p_valid_i = Signal(reset_less=True)
        nir_novn = Signal(reset_less=True)
        nirn_novn = Signal(reset_less=True)
        por_pivn = Signal(reset_less=True)
        npnn = Signal(reset_less=True)
        self.m.d.comb += [p_valid_i.eq(self.p.valid_i_test),
                     o_n_validn.eq(~self.n.valid_o),
                     n_ready_i.eq(self.n.ready_i_test),
                     nir_por.eq(n_ready_i & self.p._ready_o),
                     nir_por_n.eq(n_ready_i & ~self.p._ready_o),
                     nir_novn.eq(n_ready_i | o_n_validn),
                     nirn_novn.eq(~n_ready_i & o_n_validn),
                     npnn.eq(nir_por | nirn_novn),
                     por_pivn.eq(self.p._ready_o & ~p_valid_i)
        ]

        # store result of processing in combinatorial temporary
        self.m.d.comb += nmoperator.eq(result, self.data_r)

        # if not in stall condition, update the temporary register
        with self.m.If(self.p.ready_o): # not stalled
            self.m.d.sync += nmoperator.eq(r_data, result) # update buffer

        # data pass-through conditions
        with self.m.If(npnn):
            data_o = self._postprocess(result) # XXX TBD, does nothing right now
            self.m.d.sync += [self.n.valid_o.eq(p_valid_i), # valid if p_valid
                              nmoperator.eq(self.n.data_o, data_o), # update out
                             ]
        # buffer flush conditions (NOTE: can override data passthru conditions)
        with self.m.If(nir_por_n): # not stalled
            # Flush the [already processed] buffer to the output port.
            data_o = self._postprocess(r_data) # XXX TBD, does nothing right now
            self.m.d.sync += [self.n.valid_o.eq(1),  # reg empty
                              nmoperator.eq(self.n.data_o, data_o), # flush
                             ]
        # output ready conditions
        self.m.d.sync += self.p._ready_o.eq(nir_novn | por_pivn)

        return self.m


class MaskNoDelayCancellable(ControlBase):
    """ Mask-activated Cancellable pipeline (that does not respect "ready")

        Based on (identical behaviour to) SimpleHandshake.
        TODO: decide whether to merge *into* SimpleHandshake.

        Argument: stage.  see Stage API above

        stage-1   p.valid_i >>in   stage   n.valid_o out>>   stage+1
        stage-1   p.ready_o <<out  stage   n.ready_i <<in    stage+1
        stage-1   p.data_i  >>in   stage   n.data_o  out>>   stage+1
                              |             |
                              +--process->--^
    """
    def __init__(self, stage, maskwid, in_multi=None, stage_ctl=False):
        ControlBase.__init__(self, stage, in_multi, stage_ctl, maskwid)

    def elaborate(self, platform):
        self.m = m = ControlBase.elaborate(self, platform)

        # store result of processing in combinatorial temporary
        result = _spec(self.stage.ospec, "r_tmp")
        m.d.comb += nmoperator.eq(result, self.data_r)

        # establish if the data should be passed on.  cancellation is
        # a global signal.
        # XXX EXCEPTIONAL CIRCUMSTANCES: inspection of the data payload
        # is NOT "normal" for the Stage API.
        p_valid_i = Signal(reset_less=True)
        #print ("self.p.data_i", self.p.data_i)
        maskedout = Signal(len(self.p.mask_i), reset_less=True)
        m.d.comb += maskedout.eq(self.p.mask_i & ~self.p.stop_i)
        m.d.comb += p_valid_i.eq(maskedout.bool())

        # if idmask nonzero, mask gets passed on (and register set).
        # register is left as-is if idmask is zero, but out-mask is set to zero
        # note however: only the *uncancelled* mask bits get passed on
        m.d.sync += self.n.valid_o.eq(p_valid_i)
        m.d.sync += self.n.mask_o.eq(Mux(p_valid_i, maskedout, 0))
        with m.If(p_valid_i):
            data_o = self._postprocess(result) # XXX TBD, does nothing right now
            m.d.sync += nmoperator.eq(self.n.data_o, data_o) # update output

        # output valid if
        # input always "ready"
        #m.d.comb += self.p._ready_o.eq(self.n.ready_i_test)
        m.d.comb += self.p._ready_o.eq(Const(1))

        # always pass on stop (as combinatorial: single signal)
        m.d.comb += self.n.stop_o.eq(self.p.stop_i)

        return self.m


class MaskCancellable(ControlBase):
    """ Mask-activated Cancellable pipeline

        Arguments:

        * stage.  see Stage API above
        * maskwid - sets up cancellation capability (mask and stop).
        * in_multi
        * stage_ctl
        * dynamic - allows switching from sync to combinatorial (passthrough)
                    USE WITH CARE.  will need the entire pipe to be quiescent
                    before switching, otherwise data WILL be destroyed.

        stage-1   p.valid_i >>in   stage   n.valid_o out>>   stage+1
        stage-1   p.ready_o <<out  stage   n.ready_i <<in    stage+1
        stage-1   p.data_i  >>in   stage   n.data_o  out>>   stage+1
                              |             |
                              +--process->--^
    """
    def __init__(self, stage, maskwid, in_multi=None, stage_ctl=False,
                       dynamic=False):
        ControlBase.__init__(self, stage, in_multi, stage_ctl, maskwid)
        self.dynamic = dynamic
        if dynamic:
            self.latchmode = Signal()
        else:
            self.latchmode = Const(1)

    def elaborate(self, platform):
        self.m = m = ControlBase.elaborate(self, platform)

        mask_r = Signal(len(self.p.mask_i), reset_less=True)
        data_r = _spec(self.stage.ospec, "data_r")
        m.d.comb += nmoperator.eq(data_r, self._postprocess(self.data_r))

        with m.If(self.latchmode):
            r_busy = Signal()
            r_latch = _spec(self.stage.ospec, "r_latch")

            # establish if the data should be passed on.  cancellation is
            # a global signal.
            p_valid_i = Signal(reset_less=True)
            #print ("self.p.data_i", self.p.data_i)
            maskedout = Signal(len(self.p.mask_i), reset_less=True)
            m.d.comb += maskedout.eq(self.p.mask_i & ~self.p.stop_i)

            # establish some combinatorial temporaries
            n_ready_i = Signal(reset_less=True, name="n_i_rdy_data")
            p_valid_i_p_ready_o = Signal(reset_less=True)
            m.d.comb += [p_valid_i.eq(self.p.valid_i_test & maskedout.bool()),
                         n_ready_i.eq(self.n.ready_i_test),
                         p_valid_i_p_ready_o.eq(p_valid_i & self.p.ready_o),
            ]

            # if idmask nonzero, mask gets passed on (and register set).
            # register is left as-is if idmask is zero, but out-mask is set to
            # zero
            # note however: only the *uncancelled* mask bits get passed on
            m.d.sync += mask_r.eq(Mux(p_valid_i, maskedout, 0))
            m.d.comb += self.n.mask_o.eq(mask_r)

            # always pass on stop (as combinatorial: single signal)
            m.d.comb += self.n.stop_o.eq(self.p.stop_i)

            stor = Signal(reset_less=True)
            m.d.comb += stor.eq(p_valid_i_p_ready_o | n_ready_i)
            with m.If(stor):
                # store result of processing in combinatorial temporary
                m.d.sync += nmoperator.eq(r_latch, data_r)

            # previous valid and ready
            with m.If(p_valid_i_p_ready_o):
                m.d.sync += r_busy.eq(1)      # output valid
            # previous invalid or not ready, however next is accepting
            with m.Elif(n_ready_i):
                m.d.sync += r_busy.eq(0) # ...so set output invalid

            # output set combinatorially from latch
            m.d.comb += nmoperator.eq(self.n.data_o, r_latch)

            m.d.comb += self.n.valid_o.eq(r_busy)
            # if next is ready, so is previous
            m.d.comb += self.p._ready_o.eq(n_ready_i)

        with m.Else():
            # pass everything straight through.  p connected to n: data,
            # valid, mask, everything.  this is "effectively" just a
            # StageChain: MaskCancellable is doing "nothing" except
            # combinatorially passing everything through
            # (except now it's *dynamically selectable* whether to do that)
            m.d.comb += self.n.valid_o.eq(self.p.valid_i_test)
            m.d.comb += self.p._ready_o.eq(self.n.ready_i_test)
            m.d.comb += self.n.stop_o.eq(self.p.stop_i)
            m.d.comb += self.n.mask_o.eq(self.p.mask_i)
            m.d.comb += nmoperator.eq(self.n.data_o, data_r)

        return self.m


class SimpleHandshake(ControlBase):
    """ simple handshake control.  data and strobe signals travel in sync.
        implements the protocol used by Wishbone and AXI4.

        Argument: stage.  see Stage API above

        stage-1   p.valid_i >>in   stage   n.valid_o out>>   stage+1
        stage-1   p.ready_o <<out  stage   n.ready_i <<in    stage+1
        stage-1   p.data_i  >>in   stage   n.data_o  out>>   stage+1
                              |             |
                              +--process->--^
        Truth Table

        Inputs   Temporary  Output Data
        -------  ---------- -----  ----
        P P N N  PiV& ~NiR&  N P
        i o i o  PoR  NoV    o o
        V R R V              V R

        -------   -    -     - -
        0 0 0 0   0    0    >0 0    reg
        0 0 0 1   0    1    >1 0    reg
        0 0 1 0   0    0     0 1    process(data_i)
        0 0 1 1   0    0     0 1    process(data_i)
        -------   -    -     - -
        0 1 0 0   0    0    >0 0    reg
        0 1 0 1   0    1    >1 0    reg
        0 1 1 0   0    0     0 1    process(data_i)
        0 1 1 1   0    0     0 1    process(data_i)
        -------   -    -     - -
        1 0 0 0   0    0    >0 0    reg
        1 0 0 1   0    1    >1 0    reg
        1 0 1 0   0    0     0 1    process(data_i)
        1 0 1 1   0    0     0 1    process(data_i)
        -------   -    -     - -
        1 1 0 0   1    0     1 0    process(data_i)
        1 1 0 1   1    1     1 0    process(data_i)
        1 1 1 0   1    0     1 1    process(data_i)
        1 1 1 1   1    0     1 1    process(data_i)
        -------   -    -     - -
    """

    def elaborate(self, platform):
        self.m = m = ControlBase.elaborate(self, platform)

        r_busy = Signal()
        result = _spec(self.stage.ospec, "r_tmp")

        # establish some combinatorial temporaries
        n_ready_i = Signal(reset_less=True, name="n_i_rdy_data")
        p_valid_i_p_ready_o = Signal(reset_less=True)
        p_valid_i = Signal(reset_less=True)
        m.d.comb += [p_valid_i.eq(self.p.valid_i_test),
                     n_ready_i.eq(self.n.ready_i_test),
                     p_valid_i_p_ready_o.eq(p_valid_i & self.p.ready_o),
        ]

        # store result of processing in combinatorial temporary
        m.d.comb += nmoperator.eq(result, self.data_r)

        # previous valid and ready
        with m.If(p_valid_i_p_ready_o):
            data_o = self._postprocess(result) # XXX TBD, does nothing right now
            m.d.sync += [r_busy.eq(1),      # output valid
                         nmoperator.eq(self.n.data_o, data_o), # update output
                        ]
        # previous invalid or not ready, however next is accepting
        with m.Elif(n_ready_i):
            data_o = self._postprocess(result) # XXX TBD, does nothing right now
            m.d.sync += [nmoperator.eq(self.n.data_o, data_o)]
            # TODO: could still send data here (if there was any)
            #m.d.sync += self.n.valid_o.eq(0) # ...so set output invalid
            m.d.sync += r_busy.eq(0) # ...so set output invalid

        m.d.comb += self.n.valid_o.eq(r_busy)
        # if next is ready, so is previous
        m.d.comb += self.p._ready_o.eq(n_ready_i)

        return self.m


class UnbufferedPipeline(ControlBase):
    """ A simple pipeline stage with single-clock synchronisation
        and two-way valid/ready synchronised signalling.

        Note that a stall in one stage will result in the entire pipeline
        chain stalling.

        Also that unlike BufferedHandshake, the valid/ready signalling does NOT
        travel synchronously with the data: the valid/ready signalling
        combines in a *combinatorial* fashion.  Therefore, a long pipeline
        chain will lengthen propagation delays.

        Argument: stage.  see Stage API, above

        stage-1   p.valid_i >>in   stage   n.valid_o out>>   stage+1
        stage-1   p.ready_o <<out  stage   n.ready_i <<in    stage+1
        stage-1   p.data_i  >>in   stage   n.data_o  out>>   stage+1
                              |             |
                            r_data        result
                              |             |
                              +--process ->-+

        Attributes:
        -----------
        p.data_i : StageInput, shaped according to ispec
            The pipeline input
        p.data_o : StageOutput, shaped according to ospec
            The pipeline output
        r_data : input_shape according to ispec
            A temporary (buffered) copy of a prior (valid) input.
            This is HELD if the output is not ready.  It is updated
            SYNCHRONOUSLY.
        result: output_shape according to ospec
            The output of the combinatorial logic.  it is updated
            COMBINATORIALLY (no clock dependence).

        Truth Table

        Inputs  Temp  Output  Data
        -------   -   -----   ----
        P P N N ~NiR&  N P
        i o i o  NoV   o o
        V R R V        V R

        -------   -    - -
        0 0 0 0   0    0 1    reg
        0 0 0 1   1    1 0    reg
        0 0 1 0   0    0 1    reg
        0 0 1 1   0    0 1    reg
        -------   -    - -
        0 1 0 0   0    0 1    reg
        0 1 0 1   1    1 0    reg
        0 1 1 0   0    0 1    reg
        0 1 1 1   0    0 1    reg
        -------   -    - -
        1 0 0 0   0    1 1    reg
        1 0 0 1   1    1 0    reg
        1 0 1 0   0    1 1    reg
        1 0 1 1   0    1 1    reg
        -------   -    - -
        1 1 0 0   0    1 1    process(data_i)
        1 1 0 1   1    1 0    process(data_i)
        1 1 1 0   0    1 1    process(data_i)
        1 1 1 1   0    1 1    process(data_i)
        -------   -    - -

        Note: PoR is *NOT* involved in the above decision-making.
    """

    def elaborate(self, platform):
        self.m = m = ControlBase.elaborate(self, platform)

        data_valid = Signal() # is data valid or not
        r_data = _spec(self.stage.ospec, "r_tmp") # output type

        # some temporaries
        p_valid_i = Signal(reset_less=True)
        pv = Signal(reset_less=True)
        buf_full = Signal(reset_less=True)
        m.d.comb += p_valid_i.eq(self.p.valid_i_test)
        m.d.comb += pv.eq(self.p.valid_i & self.p.ready_o)
        m.d.comb += buf_full.eq(~self.n.ready_i_test & data_valid)

        m.d.comb += self.n.valid_o.eq(data_valid)
        m.d.comb += self.p._ready_o.eq(~data_valid | self.n.ready_i_test)
        m.d.sync += data_valid.eq(p_valid_i | buf_full)

        with m.If(pv):
            m.d.sync += nmoperator.eq(r_data, self.data_r)
        data_o = self._postprocess(r_data) # XXX TBD, does nothing right now
        m.d.comb += nmoperator.eq(self.n.data_o, data_o)

        return self.m


class UnbufferedPipeline2(ControlBase):
    """ A simple pipeline stage with single-clock synchronisation
        and two-way valid/ready synchronised signalling.

        Note that a stall in one stage will result in the entire pipeline
        chain stalling.

        Also that unlike BufferedHandshake, the valid/ready signalling does NOT
        travel synchronously with the data: the valid/ready signalling
        combines in a *combinatorial* fashion.  Therefore, a long pipeline
        chain will lengthen propagation delays.

        Argument: stage.  see Stage API, above

        stage-1   p.valid_i >>in   stage   n.valid_o out>>   stage+1
        stage-1   p.ready_o <<out  stage   n.ready_i <<in    stage+1
        stage-1   p.data_i  >>in   stage   n.data_o  out>>   stage+1
                              |             |    |
                              +- process-> buf <-+
        Attributes:
        -----------
        p.data_i : StageInput, shaped according to ispec
            The pipeline input
        p.data_o : StageOutput, shaped according to ospec
            The pipeline output
        buf : output_shape according to ospec
            A temporary (buffered) copy of a valid output
            This is HELD if the output is not ready.  It is updated
            SYNCHRONOUSLY.

        Inputs  Temp  Output Data
        -------   -   -----
        P P N N ~NiR&  N P   (buf_full)
        i o i o  NoV   o o
        V R R V        V R

        -------   -    - -
        0 0 0 0   0    0 1   process(data_i)
        0 0 0 1   1    1 0   reg (odata, unchanged)
        0 0 1 0   0    0 1   process(data_i)
        0 0 1 1   0    0 1   process(data_i)
        -------   -    - -
        0 1 0 0   0    0 1   process(data_i)
        0 1 0 1   1    1 0   reg (odata, unchanged)
        0 1 1 0   0    0 1   process(data_i)
        0 1 1 1   0    0 1   process(data_i)
        -------   -    - -
        1 0 0 0   0    1 1   process(data_i)
        1 0 0 1   1    1 0   reg (odata, unchanged)
        1 0 1 0   0    1 1   process(data_i)
        1 0 1 1   0    1 1   process(data_i)
        -------   -    - -
        1 1 0 0   0    1 1   process(data_i)
        1 1 0 1   1    1 0   reg (odata, unchanged)
        1 1 1 0   0    1 1   process(data_i)
        1 1 1 1   0    1 1   process(data_i)
        -------   -    - -

        Note: PoR is *NOT* involved in the above decision-making.
    """

    def elaborate(self, platform):
        self.m = m = ControlBase.elaborate(self, platform)

        buf_full = Signal() # is data valid or not
        buf = _spec(self.stage.ospec, "r_tmp") # output type

        # some temporaries
        p_valid_i = Signal(reset_less=True)
        m.d.comb += p_valid_i.eq(self.p.valid_i_test)

        m.d.comb += self.n.valid_o.eq(buf_full | p_valid_i)
        m.d.comb += self.p._ready_o.eq(~buf_full)
        m.d.sync += buf_full.eq(~self.n.ready_i_test & self.n.valid_o)

        data_o = Mux(buf_full, buf, self.data_r)
        data_o = self._postprocess(data_o) # XXX TBD, does nothing right now
        m.d.comb += nmoperator.eq(self.n.data_o, data_o)
        m.d.sync += nmoperator.eq(buf, self.n.data_o)

        return self.m


class PassThroughHandshake(ControlBase):
    """ A control block that delays by one clock cycle.

        Inputs   Temporary          Output Data
        -------  ------------------  ----- ----
        P P N N  PiV& PiV| NiR| pvr   N P  (pvr)
        i o i o  PoR  ~PoR ~NoV       o o
        V R R V                       V R

        -------   -    -    -   -     - -
        0 0 0 0   0    1    1   0     1 1   odata (unchanged)
        0 0 0 1   0    1    0   0     1 0   odata (unchanged)
        0 0 1 0   0    1    1   0     1 1   odata (unchanged)
        0 0 1 1   0    1    1   0     1 1   odata (unchanged)
        -------   -    -    -   -     - -
        0 1 0 0   0    0    1   0     0 1   odata (unchanged)
        0 1 0 1   0    0    0   0     0 0   odata (unchanged)
        0 1 1 0   0    0    1   0     0 1   odata (unchanged)
        0 1 1 1   0    0    1   0     0 1   odata (unchanged)
        -------   -    -    -   -     - -
        1 0 0 0   0    1    1   1     1 1   process(in)
        1 0 0 1   0    1    0   0     1 0   odata (unchanged)
        1 0 1 0   0    1    1   1     1 1   process(in)
        1 0 1 1   0    1    1   1     1 1   process(in)
        -------   -    -    -   -     - -
        1 1 0 0   1    1    1   1     1 1   process(in)
        1 1 0 1   1    1    0   0     1 0   odata (unchanged)
        1 1 1 0   1    1    1   1     1 1   process(in)
        1 1 1 1   1    1    1   1     1 1   process(in)
        -------   -    -    -   -     - -

    """

    def elaborate(self, platform):
        self.m = m = ControlBase.elaborate(self, platform)

        r_data = _spec(self.stage.ospec, "r_tmp") # output type

        # temporaries
        p_valid_i = Signal(reset_less=True)
        pvr = Signal(reset_less=True)
        m.d.comb += p_valid_i.eq(self.p.valid_i_test)
        m.d.comb += pvr.eq(p_valid_i & self.p.ready_o)

        m.d.comb += self.p.ready_o.eq(~self.n.valid_o |  self.n.ready_i_test)
        m.d.sync += self.n.valid_o.eq(p_valid_i       | ~self.p.ready_o)

        odata = Mux(pvr, self.data_r, r_data)
        m.d.sync += nmoperator.eq(r_data, odata)
        r_data = self._postprocess(r_data) # XXX TBD, does nothing right now
        m.d.comb += nmoperator.eq(self.n.data_o, r_data)

        return m


class RegisterPipeline(UnbufferedPipeline):
    """ A pipeline stage that delays by one clock cycle, creating a
        sync'd latch out of data_o and valid_o as an indirect byproduct
        of using PassThroughStage
    """
    def __init__(self, iospecfn):
        UnbufferedPipeline.__init__(self, PassThroughStage(iospecfn))


class FIFOControl(ControlBase):
    """ FIFO Control.  Uses Queue to store data, coincidentally
        happens to have same valid/ready signalling as Stage API.

        data_i -> fifo.din -> FIFO -> fifo.dout -> data_o
    """
    def __init__(self, depth, stage, in_multi=None, stage_ctl=False,
                                     fwft=True, pipe=False):
        """ FIFO Control

            * :depth: number of entries in the FIFO
            * :stage: data processing block
            * :fwft:  first word fall-thru mode (non-fwft introduces delay)
            * :pipe:  specifies pipe mode.

            when fwft = True it indicates that transfers may occur
            combinatorially through stage processing in the same clock cycle.
            This requires that the Stage be a Moore FSM:
            https://en.wikipedia.org/wiki/Moore_machine

            when fwft = False it indicates that all output signals are
            produced only from internal registers or memory, i.e. that the
            Stage is a Mealy FSM:
            https://en.wikipedia.org/wiki/Mealy_machine

            data is processed (and located) as follows:

            self.p  self.stage temp    fn temp  fn  temp  fp   self.n
            data_i->process()->result->cat->din.FIFO.dout->cat(data_o)

            yes, really: cat produces a Cat() which can be assigned to.
            this is how the FIFO gets de-catted without needing a de-cat
            function
        """
        self.fwft = fwft
        self.pipe = pipe
        self.fdepth = depth
        ControlBase.__init__(self, stage, in_multi, stage_ctl)

    def elaborate(self, platform):
        self.m = m = ControlBase.elaborate(self, platform)

        # make a FIFO with a signal of equal width to the data_o.
        (fwidth, _) = nmoperator.shape(self.n.data_o)
        fifo = Queue(fwidth, self.fdepth, fwft=self.fwft, pipe=self.pipe)
        m.submodules.fifo = fifo

        def processfn(data_i):
            # store result of processing in combinatorial temporary
            result = _spec(self.stage.ospec, "r_temp")
            m.d.comb += nmoperator.eq(result, self.process(data_i))
            return nmoperator.cat(result)

        ## prev: make the FIFO (Queue object) "look" like a PrevControl...
        m.submodules.fp = fp = PrevControl()
        fp.valid_i, fp._ready_o, fp.data_i = fifo.w_en, fifo.w_rdy, fifo.w_data
        m.d.comb += fp._connect_in(self.p, fn=processfn)

        # next: make the FIFO (Queue object) "look" like a NextControl...
        m.submodules.fn = fn = NextControl()
        fn.valid_o, fn.ready_i, fn.data_o  = fifo.r_rdy, fifo.r_en, fifo.r_data
        connections = fn._connect_out(self.n, fn=nmoperator.cat)
        valid_eq, ready_eq, data_o = connections

        # ok ok so we can't just do the ready/valid eqs straight:
        # first 2 from connections are the ready/valid, 3rd is data.
        if self.fwft:
            m.d.comb += [valid_eq, ready_eq] # combinatorial on next ready/valid
        else:
            m.d.sync += [valid_eq, ready_eq] # non-fwft mode needs sync
        data_o = self._postprocess(data_o) # XXX TBD, does nothing right now
        m.d.comb += data_o

        return m


# aka "RegStage".
class UnbufferedPipeline(FIFOControl):
    def __init__(self, stage, in_multi=None, stage_ctl=False):
        FIFOControl.__init__(self, 1, stage, in_multi, stage_ctl,
                                   fwft=True, pipe=False)

# aka "BreakReadyStage" XXX had to set fwft=True to get it to work
class PassThroughHandshake(FIFOControl):
    def __init__(self, stage, in_multi=None, stage_ctl=False):
        FIFOControl.__init__(self, 1, stage, in_multi, stage_ctl,
                                   fwft=True, pipe=True)

# this is *probably* BufferedHandshake, although test #997 now succeeds.
class BufferedHandshake(FIFOControl):
    def __init__(self, stage, in_multi=None, stage_ctl=False):
        FIFOControl.__init__(self, 2, stage, in_multi, stage_ctl,
                                   fwft=True, pipe=False)


"""
# this is *probably* SimpleHandshake (note: memory cell size=0)
class SimpleHandshake(FIFOControl):
    def __init__(self, stage, in_multi=None, stage_ctl=False):
        FIFOControl.__init__(self, 0, stage, in_multi, stage_ctl,
                                   fwft=True, pipe=False)
"""



"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+

"""

from collections.abc import Iterable
from nmigen import Mux, Signal, Cat

# XXX this already exists in nmigen._utils
# see https://bugs.libre-soc.org/show_bug.cgi?id=297
def flatten(v):
    if isinstance(v, Iterable):
        for i in v:
            yield from flatten(i)
    else:
        yield v

# tree reduction function.  operates recursively.
def treereduce(tree, op, fn):
    """treereduce: apply a map-reduce to a list.
    examples: OR-reduction of one member of a list of Records down to a
              single data point:
              treereduce(tree, operator.or_, lambda x: getattr(x, "data_o"))
    """
    #print ("treereduce", tree)
    if not isinstance(tree, list):
        return tree
    if len(tree) == 1:
        return fn(tree[0])
    if len(tree) == 2:
        return op(fn(tree[0]), fn(tree[1]))
    s = len(tree) // 2 # splitpoint
    return op(treereduce(tree[:s], op, fn),
              treereduce(tree[s:], op, fn))

# chooses assignment of 32 bit or full 64 bit depending on is_32bit
def eq32(is_32bit, dest, src):
    return [dest[0:32].eq(src[0:32]),
            dest[32:64].eq(Mux(is_32bit, 0, src[32:64]))]


# a wrapper function formerly in run_simulation that is still useful.
# Simulation.add_sync_process now only takes functions, it does not
# take generators.  so passing in arguments is no longer possible.
# with this wrapper, the following is possible:
#       sim.add_sync_process(wrap.dut(parallel_sender_number=0))
#       sim.add_sync_process(wrap.dut(parallel_sender_number=1))

def wrap(process):
    def wrapper():
        yield from process
    return wrapper


# a "rising edge" generator.  can take signals of greater than width 1

def rising_edge(m, sig):
    delay = Signal.like(sig)
    rising = Signal.like(sig)
    delay.name = "%s_dly" % sig.name
    rising.name = "%s_rise" % sig.name
    m.d.sync += delay.eq(sig) # 1 clock delay
    m.d.comb += rising.eq(sig & ~delay) # sig is hi but delay-sig is lo
    return rising


# Display function (dummy if non-existent)
# added as a patch from jeanthom
# https://gist.githubusercontent.com/jeanthom/
#           f97f5b928720d4adda9d295e8a5bc078/
#           raw/694274e0aceec993c0fc127e296b1a85b93c1b89/nmigen-display.diff
try:
    from nmigen.hdl.ast import Display
except ImportError:
    def Display(*args):
        return []


def sel(m, r, sel_bits, field_width=None, name=None, src_loc_at=0):
    """Forms a subfield from a selection of bits of the signal `r`
    ("register").

    :param m: nMigen Module for adding the wires
    :param r: signal containing the field from which to select the subfield
    :param sel_bits: bit indices of the subfield, in "MSB 0" convention,
                     from most significant to least significant. Note that
                     the indices are allowed to be non-contiguous and/or
                     out-of-order.
    :param field_width: field width. If absent, use the signal `r` own width.
    :param name: name of the generated Signal
    :param src_loc_at: in the absence of `name`, stack level in which
                       to find it

    :returns: a new Signal which gets assigned to the subfield
    """
    # find the MSB index in LSB0 numbering
    if field_width is None:
        msb = len(r) - 1
    else:
        msb = field_width - 1
    # extract the selected bits
    sig_list = []
    for idx in sel_bits:
        sig_list.append(r[msb - idx])
    # place the LSB at the front of the list,
    # since, in nMigen, Cat starts from the LSB
    sig_list.reverse()
    sel_ret = Signal(len(sig_list), name=name, src_loc_at=src_loc_at+1)
    m.d.comb += sel_ret.eq(Cat(*sig_list))
    return sel_ret



"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


"""

from nmigen import Module, Signal, Elaboratable
from nmigen.utils import log2_int

def masked(m_out, m_in, mask):
    return (m_out & ~mask) | (m_in & mask)


class Mask(Elaboratable):
    def __init__(self, sz):
        self.sz = sz
        self.shift = Signal(log2_int(sz, False)+1)
        self.mask = Signal(sz)

    def elaborate(self, platform):
        m = Module()

        for i in range(self.sz):
            with m.If(self.shift > i):
                m.d.comb += self.mask[i].eq(1)

        return m




from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()

version = '0.0.1'

install_requires = [
    'pyvcd',  # for stylish GTKWave documents - available on Pypi
]

test_requires = [
    'nose',
]

setup(
    name='libresoc-nmutil',
    version=version,
    description="A nmigen utility library",
    long_description=README + '\n\n' + NEWS,
    classifiers=[
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python :: 3",
    ],
    keywords='nmigen utilities',
    author='Luke Kenneth Casson Leighton',
    author_email='lkcl@libre-soc.org',
    url='http://git.libre-soc.org/?p=nmutil.git',
    license='LGPLv3+',
    packages=find_packages('src'),
    package_dir = {'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=test_requires,
    test_suite='nose.collector',
)



""" Stage API

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


    Associated development bugs:
    * http://bugs.libre-riscv.org/show_bug.cgi?id=148
    * http://bugs.libre-riscv.org/show_bug.cgi?id=64
    * http://bugs.libre-riscv.org/show_bug.cgi?id=57

    Stage API:
    ---------

    stage requires compliance with a strict API that may be
    implemented in several means, including as a static class.

    Stages do not HOLD data, and they definitely do not contain
    signalling (ready/valid).  They do however specify the FORMAT
    of the incoming and outgoing data, and they provide a means to
    PROCESS that data (from incoming format to outgoing format).

    Stage Blocks really should be combinatorial blocks (Moore FSMs).
    It would be ok to have input come in from sync'd sources
    (clock-driven, Mealy FSMs) however by doing so they would no longer
    be deterministic, and chaining such blocks with such side-effects
    together could result in unexpected, unpredictable, unreproduceable
    behaviour.

    So generally to be avoided, then unless you know what you are doing.
    https://en.wikipedia.org/wiki/Moore_machine
    https://en.wikipedia.org/wiki/Mealy_machine

    the methods of a stage instance must be as follows:

    * ispec() - Input data format specification.  Takes a bit of explaining.
                The requirements are: something that eventually derives from
                nmigen Value must be returned *OR* an iterator or iterable
                or sequence (list, tuple etc.) or generator must *yield*
                thing(s) that (eventually) derive from the nmigen Value class.

                Complex to state, very simple in practice:
                see test_buf_pipe.py for over 25 worked examples.

    * ospec() - Output data format specification.
                format requirements identical to ispec.

    * process(m, i) - Optional function for processing ispec-formatted data.
                returns a combinatorial block of a result that
                may be assigned to the output, by way of the "nmoperator.eq"
                function.  Note that what is returned here can be
                extremely flexible.  Even a dictionary can be returned
                as long as it has fields that match precisely with the
                Record into which its values is intended to be assigned.
                Again: see example unit tests for details.

    * setup(m, i) - Optional function for setting up submodules.
                may be used for more complex stages, to link
                the input (i) to submodules.  must take responsibility
                for adding those submodules to the module (m).
                the submodules must be combinatorial blocks and
                must have their inputs and output linked combinatorially.

    Both StageCls (for use with non-static classes) and Stage (for use
    by static classes) are abstract classes from which, for convenience
    and as a courtesy to other developers, anything conforming to the
    Stage API may *choose* to derive.  See Liskov Substitution Principle:
    https://en.wikipedia.org/wiki/Liskov_substitution_principle

    StageChain:
    ----------

    A useful combinatorial wrapper around stages that chains them together
    and then presents a Stage-API-conformant interface.  By presenting
    the same API as the stages it wraps, it can clearly be used recursively.

    StageHelper:
    ----------

    A convenience wrapper around a Stage-API-compliant "thing" which
    complies with the Stage API and provides mandatory versions of
    all the optional bits.
"""

from nmigen import Elaboratable
from abc import ABCMeta, abstractmethod
import inspect

from nmutil import nmoperator


def _spec(fn, name=None):
    """ useful function that determines if "fn" has an argument "name".
        if so, fn(name) is called otherwise fn() is called.

        means that ispec and ospec can be declared with *or without*
        a name argument.  normally it would be necessary to have
        "ispec(name=None)" to achieve the same effect.
    """
    if name is None:
        return fn()
    varnames = dict(inspect.getmembers(fn.__code__))['co_varnames']
    if 'name' in varnames:
        return fn(name=name)
    return fn()


class StageCls(metaclass=ABCMeta):
    """ Class-based "Stage" API.  requires instantiation (after derivation)

        see "Stage API" above..  Note: python does *not* require derivation
        from this class.  All that is required is that the pipelines *have*
        the functions listed in this class.  Derivation from this class
        is therefore merely a "courtesy" to maintainers.
    """
    @abstractmethod
    def ispec(self): pass       # REQUIRED
    @abstractmethod
    def ospec(self): pass       # REQUIRED
    #@abstractmethod
    #def setup(self, m, i): pass # OPTIONAL
    #@abstractmethod
    #def process(self, i): pass  # OPTIONAL


class Stage(metaclass=ABCMeta):
    """ Static "Stage" API.  does not require instantiation (after derivation)

        see "Stage API" above.  Note: python does *not* require derivation
        from this class.  All that is required is that the pipelines *have*
        the functions listed in this class.  Derivation from this class
        is therefore merely a "courtesy" to maintainers.
    """
    @staticmethod
    @abstractmethod
    def ispec(): pass

    @staticmethod
    @abstractmethod
    def ospec(): pass

    #@staticmethod
    #@abstractmethod
    #def setup(m, i): pass

    #@staticmethod
    #@abstractmethod
    #def process(i): pass


class StageHelper(Stage):
    """ a convenience wrapper around something that is Stage-API-compliant.
        (that "something" may be a static class, for example).

        StageHelper happens to also be compliant with the Stage API,
        it differs from the stage that it wraps in that all the "optional"
        functions are provided (hence the designation "convenience wrapper")
    """
    def __init__(self, stage):
        self.stage = stage
        self._ispecfn = None
        self._ospecfn = None
        if stage is not None:
            self.set_specs(self, self)

    def ospec(self, name=None):
        assert self._ospecfn is not None
        return _spec(self._ospecfn, name)

    def ispec(self, name=None):
        assert self._ispecfn is not None
        return _spec(self._ispecfn, name)

    def set_specs(self, p, n):
        """ sets up the ispecfn and ospecfn for getting input and output data
        """
        if hasattr(p, "stage"):
            p = p.stage
        if hasattr(n, "stage"):
            n = n.stage
        self._ispecfn = p.ispec
        self._ospecfn = n.ospec

    def new_specs(self, name):
        """ allocates new ispec and ospec pair
        """
        return (_spec(self.ispec, "%s_i" % name),
                _spec(self.ospec, "%s_o" % name))

    def process(self, i):
        if self.stage and hasattr(self.stage, "process"):
            return self.stage.process(i)
        return i

    def setup(self, m, i):
        if self.stage is not None and hasattr(self.stage, "setup"):
            self.stage.setup(m, i)

    def _postprocess(self, i): # XXX DISABLED
        return i # RETURNS INPUT
        if hasattr(self.stage, "postprocess"):
            return self.stage.postprocess(i)
        return i


class StageChain(StageHelper):
    """ pass in a list of stages (combinatorial blocks), and they will
        automatically be chained together via their input and output specs
        into a combinatorial chain, to create one giant combinatorial
        block.

        the end result conforms to the exact same Stage API.

        * input to this class will be the input of the first stage
        * output of first stage goes into input of second
        * output of second goes into input into third
        * ... (etc. etc.)
        * the output of this class will be the output of the last stage

        NOTE: whilst this is very similar to ControlBase.connect(), it is
        *really* important to appreciate that StageChain is pure
        combinatorial and bypasses (does not involve, at all, ready/valid
        signalling OF ANY KIND).

        ControlBase.connect on the other hand respects, connects, and uses
        ready/valid signalling.

        Arguments:

        * :chain: a chain of combinatorial blocks conforming to the Stage API
                  NOTE: StageChain.ispec and ospect have to have something
                  to return (beginning and end specs of the chain),
                  therefore the chain argument must be non-zero length

        * :specallocate: if set, new input and output data will be allocated
                         and connected (eq'd) to each chained Stage.
                         in some cases if this is not done, the nmigen warning
                         "driving from two sources, module is being flattened"
                         will be issued.

        NOTE: DO NOT use StageChain with combinatorial blocks that have
        side-effects (state-based / clock-based input) or conditional
        (inter-chain) dependencies, unless you really know what you are doing.
    """
    def __init__(self, chain, specallocate=False):
        assert len(chain) > 0, "stage chain must be non-zero length"
        self.chain = chain
        StageHelper.__init__(self, None)
        if specallocate:
            self.setup = self._sa_setup
        else:
            self.setup = self._na_setup
        self.set_specs(self.chain[0], self.chain[-1])

    def _sa_setup(self, m, i):
        for (idx, c) in enumerate(self.chain):
            if hasattr(c, "setup"):
                c.setup(m, i)               # stage may have some module stuff
            ofn = self.chain[idx].ospec     # last assignment survives
            cname = 'chainin%d' % idx
            o = _spec(ofn, cname)
            if isinstance(o, Elaboratable):
                setattr(m.submodules, cname, o)
            m.d.comb += nmoperator.eq(o, c.process(i)) # process input into "o"
            if idx == len(self.chain)-1:
                break
            ifn = self.chain[idx+1].ispec   # new input on next loop
            i = _spec(ifn, 'chainin%d' % (idx+1))
            m.d.comb += nmoperator.eq(i, o) # assign to next input
        self.o = o
        return self.o                       # last loop is the output

    def _na_setup(self, m, i):
        for (idx, c) in enumerate(self.chain):
            if hasattr(c, "setup"):
                c.setup(m, i)               # stage may have some module stuff
            i = o = c.process(i)            # store input into "o"
        self.o = o
        return self.o                       # last loop is the output

    def process(self, i):
        return self.o # conform to Stage API: return last-loop output





"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


"""

from copy import copy
# this is a POWER ISA 3.0B compatible *signed* div function
# however it is also the c, c++, rust, java *and* x86 way of doing things


def trunc_divs(n, d):
    abs_n = abs(n.to_signed_int())
    abs_d = abs(d.to_signed_int())
    print("trunc_div n", abs_n, n.to_signed_int())
    print("trunc_div d", abs_d, d.to_signed_int())
    abs_q = abs_n // abs_d
    sign_n = n.value & (1 << (n.bits - 1)) != 0
    sign_d = d.value & (1 << (d.bits - 1)) != 0
    print("trunc_div", hex(n.value), hex(d.value),
          hex(abs_n), hex(abs_d), hex(abs_q),
          sign_n, sign_d)
    res = copy(n)
    if (sign_n == sign_d):
        res.value = abs_q
        if res.value & (1 << (res.bits - 1)) != 0:
            # result should be positive but is negative
            raise OverflowError()
        return res
    mask = (1 << res.bits) - 1
    res.value = (-abs_q) & mask

    return res


# this is a POWER ISA 3.0B compatible *signed* mod / remainder function
# however it is also the c, c++, rust, java *and* x86 way of doing things
def trunc_rems(n, d):
    m = d * trunc_divs(n, d)

    # cheat - really shouldn't do this. mul returns full length
    m.bits = n.bits
    return n - m



# SPDX-License-Identifier: LGPL-2.1-or-later
# See Notices.txt for copyright information

""" Meta-class that allows a dynamic runtime parameter-selectable "mixin"

    This work is funded through NLnet under Grant 2019-02-012

The reasons why this technique is being deployed is because SimpleHandshake
needs to be dynamically replaced at the end-users' choice, without having
to duplicate dozens of classes using multiple-inheritanc "Mix-in" techniques.

It is however extremely unusual, and has been explicitly limited to this *one*
module.  DO NOT try to use this technique elsewhere, it is extremely hard to
understand (meta-class programming).

"""

from abc import ABCMeta

from nmutil.singlepipe import SimpleHandshake
from nmutil.singlepipe import MaskCancellable

import threading

# with many thanks to jsbueno on stackexchange for this one
# https://stackoverflow.com/questions/57273070/
# list post:
# http://lists.libre-riscv.org/pipermail/libre-riscv-dev/2019-July/002259.html

class Meta(ABCMeta):
    registry = {}
    recursing = threading.local()
    recursing.check = False
    mlock = threading.Lock()

    def __call__(cls, *args, **kw):
        mcls = cls.__class__
        if mcls.recursing.check:
            return super().__call__(*args, **kw)
        spec = args[0]
        base = spec.pipekls # pick up the dynamic class from PipelineSpec, HERE

        if (cls, base) not in mcls.registry:
            print ("__call__", args, kw, cls, base,
                   base.__bases__, cls.__bases__)
            mcls.registry[cls, base] = type(
                cls.__name__,
                (cls, base) + cls.__bases__[1:],
                {}
            )
        real_cls = mcls.registry[cls, base]

        with mcls.mlock:
            mcls.recursing.check = True
            instance = real_cls.__class__.__call__(real_cls, *args, **kw)
            mcls.recursing.check = False
        return instance


# Inherit from this class instead of SimpleHandshake (or other ControlBase
# derivative), and the metaclass will instead *replace* DynamicPipe -
# *at runtime* - with the class that is specified *as a parameter*
# in PipelineSpec.
#
# as explained in the list posting and in the stackexchange post, this is
# needed to avoid a MASSIVE suite of duplicated multiple-inheritance classes
# that "Mix in" SimpleHandshake (or other).
#
# unfortunately, composition does not work in this instance
# (make an *instance* of SimpleHandshake or other class and pass it in)
# due to the multiple level inheritance, and in several places
# the inheriting class needs to do some setup that the deriving class
# needs in order to function correctly.

class DynamicPipe(metaclass=Meta):
    def __init__(self, *args):
        print ("DynamicPipe init", super(), args)
        super().__init__(self, *args)


# bad hack: the DynamicPipe metaclass ends up creating an __init__ signature
# for the dynamically-derived class.  luckily, SimpleHandshake only needs
# "self" as the 1st argument (it is its own "Stage").  anything else
# could hypothetically be passed through the pspec.
class SimpleHandshakeRedir(SimpleHandshake):
    def __init__(self, mod, *args):
        print ("redir", mod, args)
        stage = self
        if args and args[0].stage:
            stage = args[0].stage
        SimpleHandshake.__init__(self, stage)


class MaskCancellableRedir(MaskCancellable):
    def __init__(self, mod, *args):
        stage = self
        maskwid = args[0].maskwid
        if args[0].stage:
            stage = args[0].stage
        print ("redir mask", mod, args, maskwid)
        MaskCancellable.__init__(self, stage, maskwid)




""" Priority Picker: optimised back-to-back PriorityEncoder and Decoder
    and MultiPriorityPicker: cascading mutually-exclusive pickers

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


    PriorityPicker: the input is N bits, the output is N bits wide and
    only one is enabled.

    MultiPriorityPicker: likewise except that there are M pickers and
    each output is guaranteed mutually exclusive.  Optionally:
    an "index" (and enable line) is also outputted.

    MultiPriorityPicker is designed for port-selection, when there are
    multiple "things" (of width N) contending for access to M "ports".
    When the M=0 "thing" requests a port, it gets allocated port 0
    (always).  However if the M=0 "thing" does *not* request a port,
    this gives the M=1 "thing" the opportunity to gain access to port 0.

    Given that N may potentially be much greater than M (16 bits wide
    where M may be e.g. only 4) we can't just ok, "ok so M=N therefore
    M=0 gets access to port 0, M=1 gets access to port 1" etc.
"""

from nmigen import Module, Signal, Cat, Elaboratable, Array, Const, Mux
from nmigen.cli import verilog, rtlil
import math


class PriorityPicker(Elaboratable):
    """ implements a priority-picker.  input: N bits, output: N bits

        * lsb_mode is for a LSB-priority picker
        * reverse_i=True is for convenient reverseal of the input bits
        * reverse_o=True is for convenient reversal of the output bits
    """
    def __init__(self, wid, lsb_mode=False, reverse_i=False, reverse_o=False):
        self.wid = wid
        # inputs
        self.lsb_mode = lsb_mode
        self.reverse_i = reverse_i
        self.reverse_o = reverse_o
        self.i = Signal(wid, reset_less=True)
        self.o = Signal(wid, reset_less=True)
        self.en_o = Signal(reset_less=True) # true if any output is true

    def elaborate(self, platform):
        m = Module()

        # works by saying, "if all previous bits were zero, we get a chance"
        res = []
        ni = Signal(self.wid, reset_less = True)
        i = list(self.i)
        if self.reverse_i:
            i.reverse()
        m.d.comb += ni.eq(~Cat(*i))
        prange = list(range(0, self.wid))
        if self.lsb_mode:
            prange.reverse()
        for n in prange:
            t = Signal(name="t%d" % n, reset_less = True)
            res.append(t)
            if n == 0:
                m.d.comb += t.eq(i[n])
            else:
                m.d.comb += t.eq(~Cat(ni[n], *i[:n]).bool())
        if self.reverse_o:
            res.reverse()
        # we like Cat(*xxx).  turn lists into concatenated bits
        m.d.comb += self.o.eq(Cat(*res))
        # useful "is any output enabled" signal
        m.d.comb += self.en_o.eq(self.o.bool()) # true if 1 input is true

        return m

    def __iter__(self):
        yield self.i
        yield self.o
        yield self.en_o

    def ports(self):
        return list(self)


class MultiPriorityPicker(Elaboratable):
    """ implements a multi-input priority picker
        Mx inputs of N bits, Mx outputs of N bits, only one is set

        Each picker masks out the one below it, such that the first
        gets top priority, the second cannot have the same bit that
        the first has set, and so on.  To do this, a "mask" accumulates
        the output from the chain, masking the input to the next chain.

        Also outputted (optional): an index for each picked "thing".
    """
    def __init__(self, wid, levels, indices=False, multiin=False):
        self.levels = levels
        self.wid = wid
        self.indices = indices
        self.multiin = multiin


        if multiin:
            # multiple inputs, multiple outputs.
            i_l = [] # array of picker outputs
            for j in range(self.levels):
                i = Signal(self.wid, name="i_%d" % j, reset_less=True)
                i_l.append(i)
            self.i = Array(i_l)
        else:
            # only the one input, but multiple (single) bit outputs
            self.i = Signal(self.wid, reset_less=True)

        # create array of (single-bit) outputs (unary)
        o_l = [] # array of picker outputs
        for j in range(self.levels):
            o = Signal(self.wid, name="o_%d" % j, reset_less=True)
            o_l.append(o)
        self.o = Array(o_l)

        # add an array of "enables"
        self.en_o = Signal(self.levels, name="en_o", reset_less=True)

        if not self.indices:
            return

        # add an array of indices
        lidx = math.ceil(math.log2(self.levels))
        idx_o = [] # store the array of indices
        for j in range(self.levels):
            i = Signal(lidx, name="idxo_%d" % j, reset_less=True)
            idx_o.append(i)
        self.idx_o = Array(idx_o)

    def elaborate(self, platform):
        m = Module()
        comb = m.d.comb

        # create Priority Pickers, accumulate their outputs and prevent
        # the next one in the chain from selecting that output bit.
        # the input from the current picker will be "masked" and connected
        # to the *next* picker on the next loop
        prev_pp = None
        p_mask = None
        pp_l = []
        for j in range(self.levels):
            if self.multiin:
                i = self.i[j]
            else:
                i = self.i
            o = self.o[j]
            pp = PriorityPicker(self.wid)
            pp_l.append(pp)
            setattr(m.submodules, "pp%d" % j, pp)
            comb += o.eq(pp.o)
            if prev_pp is None:
                comb += pp.i.eq(i)
                p_mask = Const(0, self.wid)
            else:
                mask = Signal(self.wid, name="m_%d" % j, reset_less=True)
                comb += mask.eq(prev_pp.o | p_mask) # accumulate output bits
                comb += pp.i.eq(i & ~mask)          # mask out input
                p_mask = mask
            i = pp.i # for input to next round
            prev_pp = pp

        # accumulate the enables
        en_l = []
        for j in range(self.levels):
            en_l.append(pp_l[j].en_o)
        # concat accumulated enable bits
        comb += self.en_o.eq(Cat(*en_l))

        if not self.indices:
            return m

        # for each picker enabled, pass that out and set a cascading index
        lidx = math.ceil(math.log2(self.levels))
        prev_count = None
        for j in range(self.levels):
            en_o = pp_l[j].en_o
            if prev_count is None:
                comb += self.idx_o[j].eq(0)
            else:
                count1 = Signal(lidx, name="count_%d" % j, reset_less=True)
                comb += count1.eq(prev_count + Const(1, lidx))
                comb += self.idx_o[j].eq(Mux(en_o, count1, prev_count))
            prev_count = self.idx_o[j]

        return m

    def __iter__(self):
        if self.multiin:
            yield from self.i
        else:
            yield self.i
        yield from self.o
        if not self.indices:
            return
        yield self.en_o
        yield from self.idx_o

    def ports(self):
        return list(self)


if __name__ == '__main__':
    dut = PriorityPicker(16)
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_picker.il", "w") as f:
        f.write(vl)
    dut = MultiPriorityPicker(5, 4, True)
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_multi_picker.il", "w") as f:
        f.write(vl)
    dut = MultiPriorityPicker(5, 4, False, True)
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_multi_picker_noidx.il", "w") as f:
        f.write(vl)



"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+

    Associated bugreports:
    * https://bugs.libre-soc.org/show_bug.cgi?id=538

"""

from nmigen import Elaboratable
from ieee754.pipeline import DynamicPipe
from nmutil.singlepipe import StageChain


class PipeModBase(Elaboratable):
    """PipeModBase: common code between nearly every pipeline module
    """
    def __init__(self, pspec, modname):
        self.modname = modname # use this to give a name to this module
        self.pspec = pspec
        self.i = self.ispec()
        self.o = self.ospec()

    def process(self, i):
        return self.o

    def setup(self, m, i):
        """ links module to inputs and outputs
        """
        setattr(m.submodules, self.modname, self)
        m.d.comb += self.i.eq(i)


class PipeModBaseChain(DynamicPipe):
    """PipeModBaseChain: common code between stage-chained pipes

    Links a set of combinatorial modules (get_chain) together
    and uses pspec.pipekls to dynamically select the pipeline type
    Also conforms to the Pipeline Stage API
    """
    def __init__(self, pspec):
        self.pspec = pspec
        self.chain = self.get_chain()
        super().__init__(pspec)

    def ispec(self):
        """ returns the input spec of the first module in the chain
        """
        return self.chain[0].ispec()

    def ospec(self):
        """ returns the output spec of the last module in the chain
        """
        return self.chain[-1].ospec()

    def process(self, i):
        return self.o # ... returned here (see setup comment below)

    def setup(self, m, i):
        """ links module to inputs and outputs
        """
        StageChain(self.chain).setup(m, i) # input linked here, through chain
        self.o = self.chain[-1].o # output is the last thing in the chain...
